<?php
if ( ! defined( 'FREEPAY_SESSION_COOKIE' ) ) {
	define( 'FREEPAY_SESSION_COOKIE', 'FREEPAY_SESSION' );
}
if ( ! class_exists( 'Recursive_ArrayAccess' ) ) {
	class Recursive_ArrayAccess implements ArrayAccess {
		protected $container = array();
		protected $dirty = false;

		protected function __construct( $data = array() ) {
			foreach ( $data as $key => $value ) {
				$this[ $key ] = $value;
			}
		}

		public function __clone() {
			foreach ( $this->container as $key => $value ) {
				if ( $value instanceof self ) {
					$this[ $key ] = clone $value;
				}
			}
		}

		public function toArray() {
			$data = $this->container;
			foreach ( $data as $key => $value ) {
				if ( $value instanceof self ) {
					$data[ $key ] = $value->toArray();
				}
			}

			return $data;
		}

		public function offsetExists( $offset ) {
			return isset( $this->container[ $offset ] );
		}

		public function offsetGet( $offset ) {
			return isset( $this->container[ $offset ] ) ? $this->container[ $offset ] : null;
		}

		public function offsetSet( $offset, $data ) {
			if ( is_array( $data ) ) {
				$data = new self( $data );
			}
			if ( $offset === null ) {
				$this->container[] = $data;
			} else {
				$this->container[ $offset ] = $data;
			}
			$this->dirty = true;
		}

		public function offsetUnset( $offset ) {
			unset( $this->container[ $offset ] );
			$this->dirty = true;
		}
	}
}
if ( ! class_exists( 'FREEPAY_SESSION' ) ) {
	final class FREEPAY_SESSION extends Recursive_ArrayAccess implements Iterator, Countable {
		protected $session_id;
		protected $expires;
		protected $exp_variant;
		private static $instance = false;

		public static function get_instance() {
			if ( ! self::$instance ) {
				self::$instance = new self();
			}

			return self::$instance;
		}

		protected function __construct() {
			if ( isset( $_COOKIE[ FREEPAY_SESSION_COOKIE ] ) ) {
				$cookie            = stripslashes( $_COOKIE[ FREEPAY_SESSION_COOKIE ] );
				$cookie_crumbs     = explode( '||', $cookie );
				$this->session_id  = $cookie_crumbs[0];
				$this->expires     = $cookie_crumbs[1];
				$this->exp_variant = $cookie_crumbs[2];
				if ( time() > $this->exp_variant ) {
					$this->set_expiration();
					delete_option( "FREEPAY_SESSION_expires_{$this->session_id}" );
					add_option( "FREEPAY_SESSION_expires_{$this->session_id}", $this->expires, '', 'no' );
				}
			} else {
				$this->session_id = $this->generate_id();
				$this->set_expiration();
			}
			$this->read_data();
			$this->set_cookie();
		}

		protected function set_expiration() {
			$this->exp_variant = time() + (int) apply_filters( '_FREEPAY_SESSION_expiration_variant', 24 * 60 );
			$this->expires     = time() + (int) apply_filters( '_FREEPAY_SESSION_expiration', 30 * 60 );
		}

		protected function set_cookie() {
			if ( ! headers_sent() ) {
				setcookie( FREEPAY_SESSION_COOKIE, $this->session_id . '||' . $this->expires . '||' . $this->exp_variant, $this->expires, COOKIEPATH, COOKIE_DOMAIN );
			}
		}

		protected function generate_id() {
			require_once( ABSPATH . 'wp-includes/class-phpass.php' );
			$hasher = new PasswordHash( 8, false );

			return md5( $hasher->get_random_bytes( 32 ) );
		}

		protected function read_data() {
			$this->container = get_option( "FREEPAY_SESSION_{$this->session_id}", array() );

			return $this->container;
		}

		public function write_data() {
			$option_key = "FREEPAY_SESSION_{$this->session_id}";
			if ( $this->dirty ) {
				if ( false === get_option( $option_key ) ) {
					add_option( "FREEPAY_SESSION_{$this->session_id}", $this->container, '', 'no' );
					add_option( "FREEPAY_SESSION_expires_{$this->session_id}", $this->expires, '', 'no' );
				} else {
					delete_option( "FREEPAY_SESSION_{$this->session_id}" );
					add_option( "FREEPAY_SESSION_{$this->session_id}", $this->container, '', 'no' );
				}
			}
		}

		public function json_out() {
			return json_encode( $this->container );
		}

		public function json_in( $data ) {
			$array = json_decode( $data );
			if ( is_array( $array ) ) {
				$this->container = $array;

				return true;
			}

			return false;
		}

		public function regenerate_id( $delete_old = false ) {
			if ( $delete_old ) {
				delete_option( "FREEPAY_SESSION_{$this->session_id}" );
			}
			$this->session_id = $this->generate_id();
			$this->set_cookie();
		}

		public function session_started() {
			return ! ! self::$instance;
		}

		public function cache_expiration() {
			return $this->expires;
		}

		public function reset() {
			$this->container = array();
		}

		public function current() {
			return current( $this->container );
		}

		public function key() {
			return key( $this->container );
		}

		public function next() {
			next( $this->container );
		}

		public function rewind() {
			reset( $this->container );
		}

		public function valid() {
			return $this->offsetExists( $this->key() );
		}

		public function count() {
			return count( $this->container );
		}
	}

	function _FREEPAY_SESSION_cache_expire() {
		$_FREEPAY_SESSION = FREEPAY_SESSION::get_instance();

		return $_FREEPAY_SESSION->cache_expiration();
	}

	function _FREEPAY_SESSION_commit() {
		_FREEPAY_SESSION_write_close();
	}

	function _FREEPAY_SESSION_decode( $data ) {
		$_FREEPAY_SESSION = FREEPAY_SESSION::get_instance();

		return $_FREEPAY_SESSION->json_in( $data );
	}

	function _FREEPAY_SESSION_encode() {
		$_FREEPAY_SESSION = FREEPAY_SESSION::get_instance();

		return $_FREEPAY_SESSION->json_out();
	}

	function _FREEPAY_SESSION_regenerate_id( $delete_old_session = false ) {
		$_FREEPAY_SESSION = FREEPAY_SESSION::get_instance();
		$_FREEPAY_SESSION->regenerate_id( $delete_old_session );

		return true;
	}

	function _FREEPAY_SESSION_start() {
		$_FREEPAY_SESSION = FREEPAY_SESSION::get_instance();
		do_action( '_FREEPAY_SESSION_start' );

		return $_FREEPAY_SESSION->session_started();
	}

	add_action( 'plugins_loaded', '_FREEPAY_SESSION_start' );
	function _FREEPAY_SESSION_status() {
		$_FREEPAY_SESSION = FREEPAY_SESSION::get_instance();
		if ( $_FREEPAY_SESSION->session_started() ) {
			return PHP_SESSION_ACTIVE;
		}

		return PHP_SESSION_NONE;
	}

	function _FREEPAY_SESSION_unset() {
		$_FREEPAY_SESSION = FREEPAY_SESSION::get_instance();
		$_FREEPAY_SESSION->reset();
	}

	function _FREEPAY_SESSION_write_close() {
		$_FREEPAY_SESSION = FREEPAY_SESSION::get_instance();
		$_FREEPAY_SESSION->write_data();
		do_action( '_FREEPAY_SESSION_commit' );
	}

	add_action( 'shutdown', '_FREEPAY_SESSION_write_close' );
	function _FREEPAY_SESSION_cleanup() {
		global $wpdb;
		if ( defined( 'FREEPAY_SETUP_CONFIG' ) ) {
			return;
		}
		if ( ! defined( 'FREEPAY_INSTALLING' ) ) {
			$expiration_keys  = $wpdb->get_results( "SELECT option_name, option_value FROM $wpdb->options WHERE option_name LIKE 'FREEPAY_SESSION_expires_%'" );
			$now              = time();
			$expired_sessions = array();
			foreach ( $expiration_keys as $expiration ) {
				if ( $now > intval( $expiration->option_value ) ) {
					$session_id         = substr( $expiration->option_name, 20 );
					$expired_sessions[] = $expiration->option_name;
					$expired_sessions[] = "FREEPAY_SESSION_$session_id";
				}
			}
			if ( ! empty( $expired_sessions ) ) {
				$option_names = implode( "','", $expired_sessions );
				$wpdb->query( "DELETE FROM $wpdb->options WHERE option_name IN ('$option_names')" );
			}
		}
		do_action( '_FREEPAY_SESSION_cleanup' );
	}

	add_action( '_FREEPAY_SESSION_garbage_collection', '_FREEPAY_SESSION_cleanup' );
	function _FREEPAY_SESSION_register_garbage_collection() {
		if ( ! wp_next_scheduled( '_FREEPAY_SESSION_garbage_collection' ) ) {
			wp_schedule_event( time(), 'hourly', '_FREEPAY_SESSION_garbage_collection' );
		}
	}

	add_action( 'wp', '_FREEPAY_SESSION_register_garbage_collection' );
}
?>